import asyncio
import os.path
from typing import Any, ClassVar, Literal, Optional, Type, TypeVar, Union

import httpx

from .models.link import LinkDelayed, LinkInfo, LinkInfos, LinkRedirect, LinkUnlock
from .models.magnet import (
    MagnetInstants,
    MagnetStatuses,
    MagnetStatusesDict,
    MagnetStatusesList,
    MagnetStatusesOne,
    MagnetUploadFiles,
    MagnetUploadURIs,
)
from .models.pin import PinCheck, PinGet
from .response import Response

T = TypeVar("T")


class Client:
    BASE: ClassVar[str] = "https://api.alldebrid.com"

    def __init__(self, apikey: Optional[str] = None, name: Optional[str] = "py.alldebrid"):
        self._apikey = apikey
        self.agent = name

    async def _request(self, method: str, url: str, **kwargs: Any) -> httpx.Response:
        async with httpx.AsyncClient(timeout=60 * 3) as client:
            if "params" in kwargs:
                kwargs["params"]["agent"] = self.agent
                kwargs["params"]["apikey"] = self._apikey
            else:
                kwargs["params"] = {"agent": self.agent, "apikey": self._apikey}
            if url.startswith("/"):
                url = self.BASE + url
            return await client.request(method, url=url, **kwargs)

    @staticmethod
    def _parse_obj(type_: Type[Response[T]], response: httpx.Response) -> T:
        try:
            output = type_.model_validate_json(response.text)
        except Exception:
            with open("last_error.json", "w") as f:
                f.write(response.text)
            raise
        return output.unwrap()

    # Pin auth

    async def pin_get(self):
        resp = await self._request("GET", "/v4/pin/get")
        return self._parse_obj(Response[PinGet], resp)

    async def pin_check(self, check: str, pin: str):
        resp = await self._request(
            "GET",
            "/v4/pin/check",
            params=dict(check=check, pin=pin),
        )
        data = self._parse_obj(Response[PinCheck], resp)
        if data.activated:
            self._apikey = data.apikey
        return data

    async def link_info(self, links: list[str], password: str = ""):
        resp = await self._request(
            "GET",
            "/v4/link/infos",
            params={"link[]": links, "password": password},
        )
        data = self._parse_obj(Response[LinkInfos], resp)
        return data

    async def link_redirect(self, link: str):
        resp = await self._request(
            "GET",
            "/v4/link/redirector",
            params=dict(link=link),
        )
        data = self._parse_obj(Response[LinkRedirect], resp)
        return data

    async def link_unlock(self, link: str, password: str = ""):
        resp = await self._request(
            "GET",
            "/v4/link/unlock",
            params=dict(link=link, password=password),
        )
        data = self._parse_obj(Response[LinkUnlock], resp)
        return data

    async def link_stream(self, id_: str, stream: str):
        resp = await self._request(
            "GET",
            "/v4/link/streaming",
            params=dict(id=id_, stream=stream),
        )
        data = self._parse_obj(Response[LinkInfo], resp)
        return data

    async def link_delayed(self, id_: str):
        resp = await self._request(
            "GET",
            "/v4/link/delayed",
            params=dict(id=id_),
        )
        data = self._parse_obj(Response[LinkDelayed], resp)
        return data

    async def link_download(self, link: str, password: str = ""):
        unlock = await self.link_unlock(link, password)
        while not unlock.link and unlock.delayed:
            delayed = await self.link_delayed(unlock.delayed)
            if delayed.link:
                unlock.link = delayed.link
            else:
                await asyncio.sleep(delayed.time_left + 1)

        return unlock

    async def magnet_upload(self, magnets: list[str]):
        resp = await self._request("POST", "/v4/magnet/upload", data={"magnets": magnets})
        return self._parse_obj(Response[MagnetUploadURIs], resp)

    async def magnet_upload_file(self, fnames: list[str]):
        resp = await self._request(
            "POST",
            "/v4/magnet/upload/file",
            files=[
                (
                    "files[]",
                    (
                        os.path.basename(fname),
                        open(fname, "rb"),
                        "application/x-bittorrent",
                    ),
                )
                for fname in fnames
            ],
        )
        return self._parse_obj(Response[MagnetUploadFiles], resp)

    async def magnet_upload_raw(self, *items: tuple[str, bytes]):
        resp = await self._request(
            "POST",
            "/v4/magnet/upload/file",
            files=[
                (
                    "files[]",
                    (
                        os.path.basename(fname),
                        data,
                        "application/x-bittorrent",
                    ),
                )
                for fname, data in items
            ],
        )
        return self._parse_obj(Response[MagnetUploadFiles], resp)

    async def magnet_status(
        self,
        id: Optional[int] = None,
        status: Union[
            Literal["active"],
            Literal["ready"],
            Literal["expired"],
            Literal["error"],
            None,
        ] = None,
    ):
        data: dict[str, Any] = {}
        if id is not None:
            data["id"] = id
        if status is not None:
            data["status"] = status
        resp = await self._request("GET", "/v4/magnet/status", params=data)
        result = self._parse_obj(Response[MagnetStatuses], resp)
        if isinstance(result, MagnetStatusesDict):
            return MagnetStatusesList(magnets=list(result.magnets.values()))
        if isinstance(result, MagnetStatusesOne):
            return MagnetStatusesList(magnets=[result.magnets])
        else:
            return result

    async def magnet_status_live(self, session: int, counter: int):
        pass

    async def magnet_delete(self, id: int):
        data = {"id": id}
        resp = await self._request("GET", "/v4/magnet/delete", params=data)
        return self._parse_obj(Response[dict[str, Any]], resp)

    async def magnet_instant(self, magnets: list[str]):
        resp = await self._request(
            "POST",
            "/v4/magnet/instant",
            data={"magnets[]": magnets},
        )
        return self._parse_obj(Response[MagnetInstants], resp)



from __future__ import annotations

from typing import Optional, Union

from pydantic import BaseModel

from .error import ErrorMessage


class LinkInfoError(BaseModel):
    link: str
    error: ErrorMessage


class LinkInfo(BaseModel):
    link: str
    filename: str
    size: int
    host: str
    hostDomain: str


class LinkInfos(BaseModel):
    infos: list[Union[LinkInfo, LinkInfoError]]


class LinkRedirect(BaseModel):
    links: list[str]


class LinkUnlockStream(BaseModel):
    id: str
    ext: str
    quality: str
    filesize: int
    name: str
    proto: Optional[str]
    link: Optional[str]


class LinkUnlock(BaseModel):
    id: str
    filename: str
    host: str
    hostDomain: Optional[str] = None
    filesize: int
    link: Optional[str] = None
    streams: Optional[list[LinkUnlockStream]] = None
    delayed: Optional[str] = None


class LinkStream(BaseModel):
    link: str
    filename: str
    filesize: int
    delayed: Optional[int]


class LinkDelayed(BaseModel):
    status: int
    time_left: int
    link: Optional[str]



from typing import Generic, Literal, Optional, TypeVar, Union, cast

from pydantic import BaseModel

from .models.error import ErrorMessage

T = TypeVar("T")


class ApiError(Exception):
    pass


class Response(BaseModel, Generic[T]):
    status: Union[Literal["success"], Literal["error"]]
    data: Optional[T] = None
    error: Optional[ErrorMessage] = None

    def unwrap(self) -> T:
        if self.status == "success":
            return cast(T, self.data)
        else:
            raise ApiError(self.error)



import itertools
from typing import Annotated, Any, Iterable, Optional, Union

from pydantic import BaseModel, BeforeValidator

from .error import ErrorMessage


class MagnetErrorURI(BaseModel):
    magnet: str
    error: ErrorMessage


class MagnetErrorFile(BaseModel):
    file: str
    error: ErrorMessage


class MagnetUploadURI(BaseModel):
    magnet: str
    name: str
    id: int
    hash: str
    size: int
    ready: bool


class MagnetUploadFile(BaseModel):
    file: str
    name: str
    id: int
    hash: str
    size: int
    ready: bool


class MagnetInstantURI(BaseModel):
    magnet: str
    hash: str
    instant: bool


class MagnetLinkEntryNormal(BaseModel):
    path: str
    fname: str
    size: int


class MagnetLinkEntry(BaseModel):
    n: str
    e: Optional[list["MagnetLinkEntry"]] = None
    s: Optional[int] = None

    @classmethod
    def parse(cls, v: dict[str, Any]):
        if "e" not in v:
            return MagnetLinkEntry(**v)
        v["e"] = [cls.parse(f) for f in v["e"]]
        return MagnetLinkEntry(**v)

    def walk(self, path: str) -> Iterable[MagnetLinkEntryNormal]:
        if self.e is not None:
            for entry in self.e:
                yield from entry.walk(path + self.n + "/")
        else:
            yield MagnetLinkEntryNormal(path=path, fname=self.n, size=self.s or 0)


class MagnetLink(BaseModel):
    @staticmethod
    def parse_files(x: list[Any]):
        return list(itertools.chain(*(MagnetLinkEntry.parse(f).walk("") for f in x)))

    link: str
    filename: str
    size: int
    files: Annotated[list[MagnetLinkEntryNormal], BeforeValidator(parse_files)]


class MagnetStatus(BaseModel):
    id: int
    filename: str
    size: int
    hash: str
    status: str
    statusCode: int
    downloaded: int
    uploaded: int
    seeders: int
    downloadSpeed: int
    uploadSpeed: int
    uploadDate: int
    completionDate: int
    links: list[MagnetLink]
    type: str
    notified: bool
    version: int


class MagnetUploadFiles(BaseModel):
    files: list[Union[MagnetUploadFile, MagnetErrorFile]]


class MagnetUploadURIs(BaseModel):
    magnets: list[Union[MagnetUploadURI, MagnetErrorURI]]


class MagnetInstants(BaseModel):
    magnets: list[Union[MagnetInstantURI, MagnetErrorURI]]


class MagnetStatusesList(BaseModel):
    magnets: list[MagnetStatus]


class MagnetStatusesDict(BaseModel):
    magnets: dict[str, MagnetStatus]


class MagnetStatusesOne(BaseModel):
    magnets: MagnetStatus


MagnetStatuses = MagnetStatusesDict | MagnetStatusesList | MagnetStatusesOne



from typing import Optional

from pydantic import BaseModel


class PinGet(BaseModel):
    pin: str
    check: str
    expires_in: int
    user_url: str
    base_url: str
    check_url: str


class PinCheck(BaseModel):
    activated: bool
    expires_in: int
    apikey: Optional[str]
